interface HolidayMap {
	[month: number]: {
		[day: number]: {
			name: string // metadata for debugging
			greetings: string[]
		}
	}
}

export const generateGreeting = (username: string, date: Date = new Date()): string => {
	const hGreeting = getHolidayGreeting(username, date)

	if (hGreeting) {
		const [holidayName, holidayGreeting] = hGreeting
		console.log(`[UTIL] Holiday: ${holidayName}`)

		return holidayGreeting
	}

	return getTimeBasedGreeting(username, date)
}

const getHolidayGreeting = (username: string, date: Date): [string, string] | null => {
	const holidays: HolidayMap = {
		1: {
			1: {
				name: "New Year",
				greetings: [
					`Happy New Year, ${username}!`,
					`Welcome to a fresh start, ${username}!`,
					`Here's to new beginnings, ${username}!`
				]
			}
		},
		2: {
			14: {
				name: "Valentine's Day",
				greetings: [
					`Happy Valentine's Day, ${username}!`,
					`Love is in the air, ${username}!`,
					`Wishing you a lovely day, ${username}!`
				]
			}
		},
		5: {
			1: {
				name: "May Day",
				greetings: [
					`Happy May Day, ${username}!`,
					`Spring celebrations await, ${username}!`,
					`Enjoy the festivities, ${username}!`
				]
			}
		},
		10: {
			31: {
				name: "Halloween",
				greetings: [
					`Happy Halloween, ${username}!`,
					`Spooky greetings, ${username}!`,
					`Trick or treat, ${username}?`
				]
			}
		},
		12: {
			25: {
				name: "Christmas",
				greetings: [
					`Merry Christmas, ${username}!`,
					`Happy holidays, ${username}!`,
					`Season's greetings, ${username}!`
				]
			},
			31: {
				name: "New Year's Eve",
				greetings: [
					`Happy New Year's Eve, ${username}!`,
					`Ready to ring in the new year, ${username}?`,
					`Goodbye to this year, ${username}!`
				]
			}
		}
	}

	const currentYear = date.getFullYear()
	const currentMonth = date.getMonth() + 1 // 0-indexed by default (lol)
	const currentDay = date.getDate()

	const [easterDay, easterMonth] = gaussEaster(currentYear)

	if (currentDay === easterDay && currentMonth === easterMonth) {
		return [
			"Easter",
			getRandomGreeting([
				`Happy Easter, ${username}!`,
				`Easter blessings, ${username}!`,
				`Hoppy Easter, ${username}!`
			])
		]
	}

	if (currentMonth === 6) {
		const midsummerDay = calculateMidsummer(currentYear)

		if (currentDay === midsummerDay) {
			return [
				"Midsummer",
				getRandomGreeting([
					`Happy Midsummer, ${username}!`,
					`Enjoying the longest days, ${username}?`,
					`Summer solstice greetings, ${username}!`,
					`Glad Midsommar, ${username}!`
				])
			]
		}
	}

	const holidayData = holidays[currentMonth]?.[currentDay]

	if (holidayData) {
		return [holidayData.name, getRandomGreeting(holidayData.greetings)]
	}

	return null
}

const calculateMidsummer = (year: number): number => {
	// saturday between 20th and 26th of June
	const startDate = new Date(year, 5, 20)
	const dayOfWeek = startDate.getDay()

	const daysUntilSaturday = dayOfWeek === 6 ? 0 : (6 - dayOfWeek + 7) % 7
	const resultDate = 20 + daysUntilSaturday

	if (resultDate > 26) {
		return resultDate - 7
	}

	return resultDate
}

const gaussEaster = (year: number): [day: number, month: number] => {
	// directly from Gauss's Easter algorithm
	const a = year % 19
	const b = year % 4
	const c = year % 7
	const p = Math.floor(year / 100.0)

	const q = Math.floor((13 + 8 * p) / 25.0)
	const m = Math.floor(15 - q + p - (Math.floor(p / 4) % 30))
	const n = Math.floor(4 + p - Math.floor(p / 4)) % 7
	const d = Math.floor(19 * a + m) % 30
	const e = Math.floor(2 * b + 4 * c + 6 * d + n) % 7

	const days = Math.floor(22 + d + e)

	if (d === 29 && e === 6) {
		return [19, 4]
	} else if (d === 28 && e === 6) {
		return [18, 4]
	}

	if (days > 31) {
		// jump to April
		return [days - 31, 4]
	}

	return [days, 3]
}

const getTimeBasedGreeting = (username: string, date: Date): string => {
	const hour = date.getHours()

	// early morning
	if (0 <= hour && hour < 5) {
		return getRandomGreeting([
			`Up late, ${username}?`,
			`Burning the midnight oil, ${username}?`,
			`Hello night owl, ${username}!`,
			`The stars are beautiful tonight, ${username}.`
		])
	}

	// morning
	if (5 <= hour && hour < 12) {
		return getRandomGreeting([
			`Good morning, ${username}!`,
			`Rise and shine, ${username}!`,
			`Top of the morning to you, ${username}!`,
			`Have a wonderful morning, ${username}!`
		])
	}

	// afternoon
	if (12 <= hour && hour < 17) {
		return getRandomGreeting([
			`Good afternoon, ${username}!`,
			`Hello there, ${username}!`,
			`Hope your day is going well, ${username}!`,
			`Afternoon greetings, ${username}!`
		])
	}

	// evening
	if (17 <= hour && hour < 21) {
		return getRandomGreeting([
			`Good evening, ${username}!`,
			`Hope you had a nice day, ${username}!`,
			`Evening greetings, ${username}!`,
			`Winding down for the day, ${username}?`
		])
	}

	// night
	return getRandomGreeting([
		`Good night, ${username}!`,
		`Having a pleasant evening, ${username}?`,
		`How was your day, ${username}?`,
		`Greetings at this fine hour, ${username}!`
	])
}

const getRandomGreeting = (greetings: string[]): string => {
	const index = Math.floor(Math.random() * greetings.length)
	return greetings[index]
}
